@@ -80,10 +80,14 @@ gjs_callback_trampoline_unref(GjsCallbackTrampoline *trampoline)
     if (trampoline->ref_count == 0) {
         JSContext *context;
 
-        context = gjs_runtime_get_current_context(trampoline->runtime);
+        context = gjs_runtime_get_context(trampoline->runtime);
 
-        if (!trampoline->is_vfunc)
+        if (!trampoline->is_vfunc) {
+            JS_BeginRequest(context);
             JS_RemoveValueRoot(context, &trampoline->js_function);
+            JS_EndRequest(context);
+        }
+
         g_callable_info_free_closure(trampoline->info, trampoline->closure);
         g_base_info_unref( (GIBaseInfo*) trampoline->info);
         g_free (trampoline->param_types);
@@ -176,7 +180,7 @@ gjs_callback_closure(ffi_cif *cif,
     g_assert(trampoline);
     gjs_callback_trampoline_ref(trampoline);
 
-    context = gjs_runtime_get_current_context(trampoline->runtime);
+    context = gjs_runtime_get_context(trampoline->runtime);
     JS_BeginRequest(context);
 
     n_args = g_callable_info_get_n_args(trampoline->info);
@@ -562,7 +566,7 @@ static JSBool
 gjs_invoke_c_function(JSContext      *context,
                       Function       *function,
                       JSObject       *obj, /* "this" object */
-                      uintN           js_argc,
+                      unsigned        js_argc,
                       jsval          *js_argv,
                       jsval          *js_rval)
 {
@@ -874,8 +878,6 @@ gjs_invoke_c_function(JSContext      *context,
          * separately */
     }
 
-    gjs_runtime_push_context(JS_GetRuntime(context), context);
-
     g_assert_cmpuint(c_arg_pos, ==, c_argc);
     g_assert_cmpuint(gi_arg_pos, ==, gi_argc);
 
@@ -890,8 +892,6 @@ gjs_invoke_c_function(JSContext      *context,
         return_value_p = &return_value.v_long;
     ffi_call(&(function->invoker.cif), function->invoker.native_address, return_value_p, ffi_arg_pointers);
 
-    gjs_runtime_pop_context(JS_GetRuntime(context));
-
     /* Return value and out arguments are valid only if invocation doesn't
      * return error. In arguments need to be released always.
      */
@@ -1180,7 +1180,7 @@ release:
 
 static JSBool
 function_call(JSContext *context,
-              uintN      js_argc,
+              unsigned   js_argc,
               jsval     *vp)
 {
     jsval *js_argv = JS_ARGV(context, vp);
@@ -1660,7 +1660,7 @@ JSBool
 gjs_invoke_c_function_uncached (JSContext      *context,
                                 GIFunctionInfo *info,
                                 JSObject       *obj,
-                                uintN           argc,
+                                unsigned        argc,
                                 jsval          *argv,
                                 jsval          *rval)
 {
