@@ -65,6 +65,20 @@ typedef struct {
     GClosure *closure;
 } ConnectData;
 
+typedef enum
+{
+  TOGGLE_DOWN,
+  TOGGLE_UP,
+} ToggleDirection;
+
+typedef struct
+{
+    JSContext       *context;
+    GObject         *gobj;
+    ToggleDirection  direction;
+    guint            needs_unref : 1;
+} ToggleRefNotifyOperation;
+
 enum {
     PROP_0,
     PROP_JS_CONTEXT,
@@ -73,6 +87,8 @@ enum {
 };
 
 static struct JSClass gjs_object_instance_class;
+static GThread *gjs_eval_thread;
+static volatile gint pending_idle_toggles;
 
 GJS_DEFINE_PRIV_FROM_JS(ObjectInstance, gjs_object_instance_class)
 
@@ -128,6 +144,26 @@ gjs_object_priv_quark (void)
     return val;
 }
 
+static GQuark
+gjs_toggle_down_quark (void)
+{
+    static GQuark val = 0;
+    if (G_UNLIKELY (!val))
+        val = g_quark_from_static_string ("gjs::toggle-down-quark");
+
+    return val;
+}
+
+static GQuark
+gjs_toggle_up_quark (void)
+{
+    static GQuark val = 0;
+    if (G_UNLIKELY (!val))
+        val = g_quark_from_static_string ("gjs::toggle-up-quark");
+
+    return val;
+}
+
 /* Plain g_type_query fails and leaves @query uninitialized for
    dynamic types.
    See https://bugzilla.gnome.org/show_bug.cgi?id=687184 and
@@ -765,6 +801,232 @@ gobj_no_longer_kept_alive_func(JSObject *obj,
     priv->keep_alive = NULL;
 }
 
+static GQuark
+get_qdata_key_for_toggle_direction(ToggleDirection direction)
+{
+    GQuark quark;
+
+    switch (direction) {
+        case TOGGLE_UP:
+            quark = gjs_toggle_up_quark();
+            break;
+        case TOGGLE_DOWN:
+            quark = gjs_toggle_down_quark();
+            break;
+        default:
+            g_assert_not_reached();
+    }
+
+    return quark;
+}
+
+static gboolean
+clear_toggle_idle_source(GObject          *gobj,
+                         ToggleDirection   direction)
+{
+    GQuark qdata_key;
+
+    qdata_key = get_qdata_key_for_toggle_direction(direction);
+
+    return g_object_steal_qdata(gobj, qdata_key) != NULL;
+}
+
+static gboolean
+toggle_idle_source_is_queued(GObject          *gobj,
+                             ToggleDirection   direction)
+{
+    GQuark qdata_key;
+
+    qdata_key = get_qdata_key_for_toggle_direction(direction);
+
+    return g_object_get_qdata(gobj, qdata_key) != NULL;
+}
+
+static gboolean
+cancel_toggle_idle(GObject         *gobj,
+                   ToggleDirection  direction)
+{
+    GQuark qdata_key;
+    GSource *source;
+
+    qdata_key = get_qdata_key_for_toggle_direction(direction);
+
+    source = g_object_steal_qdata(gobj, qdata_key);
+
+    if (source)
+        g_source_destroy(source);
+
+    return source != 0;
+}
+
+static void
+handle_toggle_down(JSContext *context,
+                   GObject   *gobj)
+{
+    ObjectInstance *priv;
+    JSObject *obj;
+
+    obj = peek_js_obj(context, gobj);
+
+    priv = priv_from_js(context, obj);
+
+    gjs_debug_lifecycle(GJS_DEBUG_GOBJECT,
+                        "Toggle notify gobj %p obj %p is_last_ref TRUE keep-alive %p",
+                        gobj, obj, priv->keep_alive);
+
+    /* Change to weak ref so the wrapper-wrappee pair can be
+     * collected by the GC
+     */
+    if (priv->keep_alive != NULL) {
+        gjs_debug_lifecycle(GJS_DEBUG_GOBJECT, "Removing object from keep alive");
+        gjs_keep_alive_remove_child(context, priv->keep_alive,
+                                    gobj_no_longer_kept_alive_func,
+                                    obj,
+                                    priv);
+        priv->keep_alive = NULL;
+    }
+}
+
+static void
+handle_toggle_up(JSContext *context,
+                 GObject   *gobj,
+                 gboolean   gc_already_blocked)
+{
+    ObjectInstance *priv;
+    JSObject *obj;
+
+    /* We need to root the JSObject associated with the passed in GObject so it
+     * doesn't get garbage collected (and lose any associated javascript state
+     * such as custom properties).
+     *
+     * Note it's possible that the garbage collector is running in a secondary
+     * thread right now. If it is, we need to wait for it to finish, then block
+     * it from starting again while we root the object. After it's blocked we need
+     * to check if the associated JSObject was reaped. If it was we need to
+     * abort mission.
+     */
+    if (!gc_already_blocked)
+        gjs_block_gc();
+
+    obj = peek_js_obj(context, gobj);
+
+    if (!obj) {
+        /* Object already GC'd */
+        goto out;
+    }
+
+    priv = priv_from_js(context, obj);
+
+    gjs_debug_lifecycle(GJS_DEBUG_GOBJECT,
+                        "Toggle notify gobj %p obj %p is_last_ref FALSEd keep-alive %p",
+                        gobj, obj, priv->keep_alive);
+
+    /* Change to strong ref so the wrappee keeps the wrapper alive
+     * in case the wrapper has data in it that the app cares about
+     */
+    if (priv->keep_alive == NULL) {
+        gjs_debug_lifecycle(GJS_DEBUG_GOBJECT, "Adding object to keep alive");
+        priv->keep_alive = gjs_keep_alive_get_for_import_global(context);
+        gjs_keep_alive_add_child(context, priv->keep_alive,
+                                 gobj_no_longer_kept_alive_func,
+                                 obj,
+                                 priv);
+    }
+
+out:
+    if (!gc_already_blocked)
+        gjs_unblock_gc();
+}
+
+static gboolean
+idle_handle_toggle(gpointer data)
+{
+    ToggleRefNotifyOperation *operation = data;
+
+    if (!clear_toggle_idle_source(operation->gobj, operation->direction)) {
+        /* Already cleared, the JSObject is going away, abort mission */
+        goto out;
+    }
+
+    switch (operation->direction) {
+        case TOGGLE_UP:
+            handle_toggle_up(operation->context, operation->gobj, FALSE);
+            break;
+        case TOGGLE_DOWN:
+            handle_toggle_down(operation->context, operation->gobj);
+            break;
+        default:
+            g_assert_not_reached();
+    }
+
+out:
+    return FALSE;
+}
+
+static void
+toggle_ref_notify_operation_free(ToggleRefNotifyOperation *operation)
+{
+    if (operation->needs_unref)
+        g_object_unref (operation->gobj);
+    g_slice_free(ToggleRefNotifyOperation, operation);
+    g_atomic_int_add(&pending_idle_toggles, -1);
+}
+
+static void
+queue_toggle_idle(GObject         *gobj,
+                  JSContext       *context,
+                  ToggleDirection  direction)
+{
+    ToggleRefNotifyOperation *operation;
+    GQuark qdata_key;
+    GSource *source;
+
+    operation = g_slice_new0(ToggleRefNotifyOperation);
+    operation->context = context;
+    operation->direction = direction;
+
+    switch (direction) {
+        case TOGGLE_UP:
+            /* If we're toggling up we take a reference to the object now,
+             * so it won't toggle down before we process it. This ensures we
+             * only ever have at most two toggle notifications queued.
+             * (either only up, or down-up)
+             */
+            operation->gobj = g_object_ref(gobj);
+            operation->needs_unref = TRUE;
+            break;
+        case TOGGLE_DOWN:
+            /* If we're toggling down, we don't need to take a reference since
+             * the associated JSObject already has one, and that JSObject won't
+             * get finalized until we've completed toggling (since it's rooted,
+             * until we unroot it when we dispatch the toggle down idle).
+             *
+             * Taking a reference now would be bad anyway, since it would force
+             * the object to toggle back up again.
+             */
+            operation->gobj = gobj;
+            break;
+        default:
+            g_assert_not_reached();
+    }
+
+    qdata_key = get_qdata_key_for_toggle_direction(direction);
+
+    source = g_idle_source_new();
+    g_source_set_priority(source, G_PRIORITY_HIGH);
+    g_source_set_callback(source,
+                          idle_handle_toggle,
+                          operation,
+                          (GDestroyNotify) toggle_ref_notify_operation_free);
+
+    g_atomic_int_inc(&pending_idle_toggles);
+    g_object_set_qdata (gobj, qdata_key, source);
+    g_source_attach (source, NULL);
+
+    /* object qdata is piggy-backing off the main loop's ref of the source */
+    g_source_unref (source);
+}
+
 static void
 wrapped_gobj_toggle_notify(gpointer      data,
                            GObject      *gobj,
@@ -772,8 +1034,8 @@ wrapped_gobj_toggle_notify(gpointer      data,
 {
     JSRuntime *runtime;
     JSContext *context;
-    JSObject *obj;
-    ObjectInstance *priv;
+    gboolean gc_blocked = FALSE;
+    gboolean toggle_up_queued, toggle_down_queued;
 
     runtime = data;
 
@@ -787,41 +1049,70 @@ wrapped_gobj_toggle_notify(gpointer      data,
     if (!context)
         return;
 
-    obj = peek_js_obj(context, gobj);
+    /* We only want to touch javascript from one thread.
+     * If we're not in that thread, then we need to defer processing
+     * to it. We also don't want to touch javascript if a GC is going
+     * on in the same thread as us.
+     *
+     * Defer to idle in those cases, and in the case where an idle
+     * is already queued (to maintain ordering constraints) but handle
+     * the toggle notify directly when we can (for efficiency reasons)
+     */
+    if (gjs_eval_thread == g_thread_self())
+        gc_blocked = gjs_try_block_gc();
 
-    g_assert(obj != NULL);
-
-    priv = priv_from_js(context, obj);
-
-    gjs_debug_lifecycle(GJS_DEBUG_GOBJECT,
-                        "Toggle notify gobj %p obj %p is_last_ref %d keep-alive %p",
-                        gobj, obj, is_last_ref, priv->keep_alive);
+    toggle_up_queued = toggle_idle_source_is_queued(gobj, TOGGLE_UP);
+    toggle_down_queued = toggle_idle_source_is_queued(gobj, TOGGLE_DOWN);
 
     if (is_last_ref) {
-        /* Change to weak ref so the wrapper-wrappee pair can be
-         * collected by the GC
+        /* We've transitions from 2 -> 1 references,
+         * The JSObject is rooted and we need to unroot it so it
+         * can be garbage collected
          */
-        if (priv->keep_alive != NULL) {
-            gjs_debug_lifecycle(GJS_DEBUG_GOBJECT, "Removing object from keep alive");
-            gjs_keep_alive_remove_child(context, priv->keep_alive,
-                                        gobj_no_longer_kept_alive_func,
-                                        obj,
-                                        priv);
-            priv->keep_alive = NULL;
+        if (gc_blocked) {
+            if (G_UNLIKELY (toggle_up_queued || toggle_down_queued)) {
+                g_error("toggling down object %s that's already queued to toggle %s\n",
+                        G_OBJECT_TYPE_NAME(gobj),
+                        toggle_up_queued && toggle_down_queued? "up and down" :
+                        toggle_up_queued? "up" : "down");
+            }
+
+            handle_toggle_down(context, gobj);
+        } else {
+            queue_toggle_idle(gobj, context, TOGGLE_DOWN);
         }
     } else {
-        /* Change to strong ref so the wrappee keeps the wrapper alive
-         * in case the wrapper has data in it that the app cares about
+        /* We've transitioned from 1 -> 2 references.
+         *
+         * The JSObject associated with the gobject is not rooted,
+         * but it needs to be. We'll root it.
          */
-        if (priv->keep_alive == NULL) {
-            gjs_debug_lifecycle(GJS_DEBUG_GOBJECT, "Adding object to keep alive");
-            priv->keep_alive = gjs_keep_alive_get_for_import_global(context);
-            gjs_keep_alive_add_child(context, priv->keep_alive,
-                                     gobj_no_longer_kept_alive_func,
-                                     obj,
-                                     priv);
+        if (gc_blocked && !toggle_down_queued) {
+            if (G_UNLIKELY (toggle_up_queued)) {
+                g_error("toggling up object %s that's already queued to toggle up\n",
+                        G_OBJECT_TYPE_NAME(gobj));
+            }
+
+            handle_toggle_up(context, gobj, gc_blocked);
+        } else {
+            queue_toggle_idle(gobj, context, TOGGLE_UP);
         }
     }
+
+    if (gc_blocked)
+        gjs_unblock_gc();
+}
+
+/* At shutdown, we need to ensure we've cleared the context of any
+ * pending toggle references.
+ */
+void
+gjs_object_process_pending_toggles (void)
+{
+    while (g_main_context_pending (NULL) &&
+           g_atomic_int_get (&pending_idle_toggles) > 0) {
+        g_main_context_iteration (NULL, FALSE);
+    }
 }
 
 static ObjectInstance *
@@ -1062,6 +1353,15 @@ object_instance_finalize(JSContext *context,
                     priv->info ? g_base_info_get_namespace((GIBaseInfo*) priv->info) : "",
                     priv->info ? g_base_info_get_name((GIBaseInfo*) priv->info) : g_type_name(priv->gtype));
         }
+
+        cancel_toggle_idle(priv->gobj, TOGGLE_UP);
+
+        if (G_UNLIKELY (cancel_toggle_idle(priv->gobj, TOGGLE_DOWN))) {
+            g_error("Finalizing proxy for an object that's scheduled to be unrooted: %s.%s\n",
+                    priv->info ? g_base_info_get_namespace((GIBaseInfo*) priv->info) : "",
+                    priv->info ? g_base_info_get_name((GIBaseInfo*) priv->info) : g_type_name(priv->gtype));
+        }
+
         set_js_obj(context, priv->gobj, NULL);
         g_object_remove_toggle_ref(priv->gobj, wrapped_gobj_toggle_notify,
                                    JS_GetRuntime(context));
@@ -2131,6 +2431,7 @@ gjs_object_class_init(GObjectClass *class,
                                                            "JSObject",
                                                            "The JSObject wrapping this GObject",
                                                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS));
+    gjs_eval_thread = g_thread_self();
 }
 
 static inline void
@@ -2172,6 +2473,7 @@ gjs_register_type(JSContext *cx,
     };
     guint32 i, n_interfaces;
     GType *iface_types;
+    JSBool retval = JS_FALSE;
 
     JS_BeginRequest(cx);
 
@@ -2180,21 +2482,21 @@ gjs_register_type(JSContext *cx,
                         "parent", &parent,
                         "name", &name,
                         "interfaces", &interfaces))
-        return JS_FALSE;
+        goto out;
 
     if (!parent)
-        return JS_FALSE;
+        goto out;
 
     if (!do_base_typecheck(cx, parent, JS_TRUE))
-        return JS_FALSE;
+        goto out;
 
     if (!JS_IsArrayObject(cx, interfaces)) {
         gjs_throw(cx, "Invalid parameter interfaces (expected Array)");
-        return JS_FALSE;
+        goto out;
     }
 
     if (!JS_GetArrayLength(cx, interfaces, &n_interfaces))
-        return JS_FALSE;
+        goto out;
 
     iface_types = g_alloca(sizeof(GType) * n_interfaces);
 
@@ -2205,13 +2507,13 @@ gjs_register_type(JSContext *cx,
         GType iface_type;
 
         if (!JS_GetElement(cx, interfaces, i, &iface_val))
-            return JS_FALSE;
+            goto out;
 
         if (!JSVAL_IS_OBJECT(iface_val) ||
             ((iface_type = gjs_gtype_get_actual_gtype(cx, JSVAL_TO_OBJECT(iface_val)))
              == G_TYPE_INVALID)) {
             gjs_throw(cx, "Invalid parameter interfaces (element %d was not a GType)", i);
-            return JS_FALSE;
+            goto out;
         }
 
         iface_types[i] = iface_type;
@@ -2219,7 +2521,7 @@ gjs_register_type(JSContext *cx,
 
     if (g_type_from_name(name) != G_TYPE_INVALID) {
         gjs_throw (cx, "Type name %s is already registered", name);
-        return JS_FALSE;
+        goto out;
     }
 
     parent_priv = priv_from_js(cx, parent);
@@ -2232,7 +2534,7 @@ gjs_register_type(JSContext *cx,
     g_type_query_dynamic_safe(parent_type, &query);
     if (G_UNLIKELY (query.type == 0)) {
         gjs_throw (cx, "Cannot inherit from a non-gjs dynamic type [bug 687184]");
-        return JS_FALSE;
+        goto out;
     }
 
     type_info.class_size = query.class_size;
@@ -2254,13 +2556,16 @@ gjs_register_type(JSContext *cx,
 
     /* create a custom JSClass */
     if (!gjs_define_object_class(cx, NULL, instance_type, &constructor, NULL))
-        return JS_FALSE;
+        goto out;
 
     JS_SET_RVAL(cx, vp, OBJECT_TO_JSVAL(constructor));
 
+    retval = JS_TRUE;
+
+out:
     JS_EndRequest(cx);
 
-    return JS_TRUE;
+    return retval;
 }
 
 static JSBool
