@@ -32,6 +32,7 @@
 #include "compat.h"
 
 #include "gi.h"
+#include "gi/object.h"
 
 #include <modules/modules.h>
 
@@ -98,6 +99,7 @@ enum {
 };
 
 
+static GMutex gc_idle_lock;
 static GMutex contexts_lock;
 static GList *all_contexts = NULL;
 
@@ -382,6 +384,16 @@ gjs_context_dispose(GObject *object)
         gjs_debug(GJS_DEBUG_CONTEXT,
                   "Destroying JS context");
 
+        gjs_object_process_pending_toggles();
+
+        /* Do a full GC here before tearing down, since once we do
+         * that we may not have the JS_GetPrivate() to access the
+         * context
+         */
+        JS_GC(js_context->context);
+
+        gjs_object_process_pending_toggles();
+
         JS_DestroyContext(js_context->context);
         js_context->context = NULL;
     }
@@ -664,8 +676,7 @@ gjs_context_constructor (GType                  type,
 
     js_context->profiler = gjs_profiler_new(js_context->runtime);
 
-    if (js_context->gc_notifications_enabled)
-        JS_SetGCCallback(js_context->context, gjs_on_context_gc);
+    JS_SetGCCallback(js_context->context, gjs_on_context_gc);
 
     JS_EndRequest(js_context->context);
 
@@ -878,8 +889,10 @@ gjs_context_idle_emit_gc (gpointer data)
 {
     GjsContext *gjs_context = data;
 
+    g_mutex_lock(&gc_idle_lock);
     gjs_context->idle_emit_gc_id = 0;
-    
+    g_mutex_unlock(&gc_idle_lock);
+
     g_signal_emit (gjs_context, signals[SIGNAL_GC], 0);
     
     return FALSE;
@@ -891,11 +904,24 @@ gjs_on_context_gc (JSContext *cx,
 {
     GjsContext *gjs_context = JS_GetContextPrivate(cx);
 
-    if (status == JSGC_END) {
-        if (gjs_context->idle_emit_gc_id == 0)
-            gjs_context->idle_emit_gc_id = g_idle_add (gjs_context_idle_emit_gc, gjs_context);
+    switch (status) {
+        case JSGC_BEGIN:
+            gjs_enter_gc();
+            break;
+        case JSGC_END:
+            gjs_leave_gc();
+            if (gjs_context->gc_notifications_enabled) {
+                g_mutex_lock(&gc_idle_lock);
+                if (gjs_context->idle_emit_gc_id == 0)
+                    gjs_context->idle_emit_gc_id = g_idle_add (gjs_context_idle_emit_gc, gjs_context);
+                g_mutex_unlock(&gc_idle_lock);
+            }
+        break;
+
+        default:
+        break;
     }
-    
+
     return TRUE;
 }
 
