@@ -35,6 +35,8 @@
 #include <string.h>
 #include <math.h>
 
+static GMutex gc_lock;
+
 GQuark
 gjs_util_error_quark (void)
 {
@@ -242,8 +244,10 @@ gjs_define_string_array(JSContext   *context,
 
     JS_BeginRequest(context);
 
-    if (!JS_EnterLocalRootScope(context))
+    if (!JS_EnterLocalRootScope(context)) {
+        JS_EndRequest(context);
         return JS_FALSE;
+    }
 
     if (array_length == -1)
         array_length = g_strv_length((char**)array_values);
@@ -353,13 +357,16 @@ gjs_value_debug_string(JSContext      *context,
                 str = JS_NewStringCopyZ(context, klass->name);
                 JS_ClearPendingException(context);
                 if (str == NULL) {
+                    JS_EndRequest(context);
                     return g_strdup("[out of memory copying class name]");
                 }
             } else {
                 gjs_log_exception(context, NULL);
+                JS_EndRequest(context);
                 return g_strdup("[unknown object]");
             }
         } else {
+            JS_EndRequest(context);
             return g_strdup("[unknown non-object]");
         }
     }
@@ -811,8 +818,10 @@ gjs_date_from_time_t (JSContext *context, time_t time)
 
     JS_BeginRequest(context);
 
-    if (!JS_EnterLocalRootScope(context))
+    if (!JS_EnterLocalRootScope(context)) {
+        JS_EndRequest(context);
         return JSVAL_VOID;
+    }
 
     if (!JS_GetClassObject(context, JS_GetGlobalObject(context), JSProto_Date,
                            &date_constructor))
@@ -1197,3 +1206,33 @@ gjs_maybe_gc (JSContext *context)
     }
 #endif
 }
+
+void
+gjs_enter_gc(void)
+{
+    g_mutex_lock(&gc_lock);
+}
+
+void
+gjs_leave_gc(void)
+{
+    g_mutex_unlock(&gc_lock);
+}
+
+gboolean
+gjs_try_block_gc(void)
+{
+    return g_mutex_trylock(&gc_lock);
+}
+
+void
+gjs_block_gc(void)
+{
+    g_mutex_lock(&gc_lock);
+}
+
+void
+gjs_unblock_gc(void)
+{
+    g_mutex_unlock(&gc_lock);
+}
